import re
from typing import Any, Dict, Union, List

def interpolate_variables(text: str, variables: Dict[str, Union[bool, int, str]]) -> str:
    def replace_match(match):
        key = match.group(1).strip()
        return str(variables.get(key, match.group(0)))

    return re.sub(r"{\s*(\w+)\s*}", replace_match, text)

def interpolate_variables_recursively(obj: Dict[str, Any], variables: Dict[str, Union[bool, int, str]]) -> Dict[str, Any]:
    def process_value(value: Any) -> Any:
        if isinstance(value, str):
            return interpolate_variables(value, variables)
        elif isinstance(value, list):
            return [process_value(v) for v in value]
        elif isinstance(value, dict):
            return process_obj(value)
        else:
            return value

    def process_obj(object: Dict[str, Any]) -> Dict[str, Any]:
        return {key: process_value(value) for key, value in object.items()}

    return process_value(obj)



import requests
from typing import Union, Dict, Any
from .classes import Prompt
from pezzo.config import pezzo_config

class Pezzo:
    def __init__(self):
        self.options = pezzo_config
        
    def get_prompt(self, prompt_name: str) -> Dict[str, Prompt]:
        url = f"{self.options.server_url}/api/prompts/v2/deployment"
        params = {
            "name": prompt_name,
            "environmentName": self.options.environment
        }
        headers = {
            "Content-Type": "application/json",
            "x-pezzo-api-key": self.options.api_key
        }
        response = requests.get(url, params=params, headers=headers)
        data = response.json()
        if not response.ok:
            raise Exception(data.get("message", f"Error fetching prompt {prompt_name} for environment {self.options.environment} ({data.get('statusCode')})."))

        pezzoPrompt = Prompt(
            metadata={
                "promptId": data.get("promptId"),
                "promptVersionSha": data.get("promptVersionSha"),
                "type": data.get("type"),
            },
            settings=data.get("settings"),
            content=data.get("content")
        )
        return pezzoPrompt

    def report_prompt_execution(self, dto: Dict[str, Any]):
        url = f"{self.options.server_url}/api/reporting/v2/request"
        headers = {
            "Content-Type": "application/json",
            "x-pezzo-api-key": self.options.api_key,
            "x-pezzo-project-id": self.options.project_id
        }

        response = requests.post(url, headers=headers, json=dto)
        if not response.ok:
            print(f"Could not report prompt execution: {response.json()}")

    def fetch_cached_request(self, request: Dict[str, Any]) -> Union[Dict[str, Any], None]:
        url = f"{self.options.server_url}/api/cache/v1/request/retrieve"
        headers = {
            "Content-Type": "application/json",
            "x-pezzo-api-key": self.options.api_key,
            "x-pezzo-project-id": self.options.project_id
        }
        response = requests.post(url, headers=headers, json={"request": request})
        if not response.ok:
            print(f"Could not fetch request from cache: {response.json()}")
            return None

        return response.json()

    def cache_request(self, request: Dict[str, Any], _response: Dict[str, Any]):
        url = f"{self.options.server_url}/api/cache/v1/request/save"
        headers = {
            "Content-Type": "application/json",
            "x-pezzo-api-key": self.options.api_key,
            "x-pezzo-project-id": self.options.project_id
        }
        response = requests.post(url, headers=headers, json={"request": request, "response": _response})
        if not response.ok:
            print(f"Could not cache request: {response.json()}")

pezzo = Pezzo()


import os
from datetime import datetime, timezone
from typing import Any, Dict, List, TypeVar, Tuple

TArgs = TypeVar('TArgs', bound=Tuple)  # Simulating 'TArgs extends unknown[]'

# Instead of interfaces, Python uses type hinting with dictionaries.
PezzoInjectedContext = Dict[str, Any]
PezzoExtendedArgs = TypeVar('PezzoExtendedArgs', bound=Tuple)  # Simulating 'PezzoExtendedArgs<TArgs>'

class ExtractedPezzoFromArgsResult:
    def __init__(self, pezzo: PezzoInjectedContext, original_args: TArgs):
        self.pezzo = pezzo
        self.original_args = original_args

def extract_pezzo_from_args(args: PezzoExtendedArgs) -> ExtractedPezzoFromArgsResult:
    pezzo = args[0].get('pezzo', {})
    original_args_0 = {k: v for k, v in args[0].items() if k != 'pezzo'}
    original_args = (original_args_0, *args[1:])

    return ExtractedPezzoFromArgsResult(pezzo, original_args)

T = TypeVar('T', bound=Dict[str, Any])

def merge(*args: T) -> T:
    result = {}
    for arg in args:
        result.update(arg)
    return result

def get_client_version() -> str:
    current_path = os.path.dirname(os.path.realpath(__file__))
    version_path = os.path.join(current_path, "../version.txt")
    with open(version_path, "r") as f:
        return f.read().strip()
    
def get_timestamp() -> str:
    timestamp = datetime.now(timezone.utc).isoformat(timespec='milliseconds').replace('+00:00', 'Z')
    return timestamp


from dataclasses import dataclass
from typing import Any, Optional, Union, List, Literal

@dataclass
class PromptMetadata:
    promptId: str
    promptVersionSha: str
    type: Literal['Prompt', 'Chat']
    isTestPrompt: Optional[bool] = None

@dataclass
class ChatPromptContentMessage:
    role: Literal['user', 'assistant']
    content: str

@dataclass
class PromptChatContent:
    messages: List["ChatPromptContentMessage"]

@dataclass
class PromptBasicContent:
    prompt: str

@dataclass
class Prompt:
    metadata: PromptMetadata
    settings: Any
    content: Union["PromptBasicContent", "PromptChatContent"]


from typing import Optional
import os

class PezzoConfig:
    _api_key: Optional[str]
    _project_id: Optional[str]
    _environment: Optional[str]
    _server_url: Optional[str]

    def __init__(self,
                 api_key: Optional[str] = None,
                 project_id: Optional[str] = None,
                 environment: Optional[str] = None,
                 server_url: Optional[str] = None
                 ):
        self._api_key = api_key
        self._project_id = project_id
        self._environment = environment
        self._server_url = server_url

    @property
    def api_key(self) -> Optional[str]:
        if (self._api_key is None):
            return os.environ.get("PEZZO_API_KEY")

        return self._api_key

    @api_key.setter
    def api_key(self, value: Optional[str]):
        self._api_key = value

    @property
    def project_id(self) -> Optional[str]:
        if (self._project_id is None):
            return os.environ.get("PEZZO_PROJECT_ID")

        return self._project_id

    @project_id.setter
    def project_id(self, value: Optional[str]):
        self._project_id = value

    @property
    def environment(self) -> Optional[str]:
        if (self._environment is None):
            return os.environ.get("PEZZO_ENVIRONMENT", "Production")
        return self._environment

    @environment.setter
    def environment(self, value: Optional[str]):
        self._environment = value

    @property
    def server_url(self) -> Optional[str]:
        if (self._server_url is None):
            return os.environ.get("PEZZO_SERVER_URL", "https://api.pezzo.ai")
        return self._server_url

    @server_url.setter
    def server_url(self, value: Optional[str]):
        self._server_url = value



pezzo_config = PezzoConfig()


import os
import openai
from pezzo.client import pezzo
from pezzo.utils.interpolate_variables import interpolate_variables_recursively
from pezzo.utils.helpers import merge, get_client_version, get_timestamp

class PezzoOpenAI:
    def __init__(self):
        openai.api_key = os.environ.get('OPENAI_API_KEY')

        if not openai.api_key:
            raise Exception('OPENAI_API_KEY is not set')

        self.openai = openai
        self.ChatCompletion = ChatCompletion(pezzo, self.openai)

class ChatCompletion:
    def __init__(self, pezzo, openai):
        self.pezzo = pezzo
        self.openai = openai

    def create(self, **args):
        pezzo_prompt = args.get('pezzo_prompt', None)
        pezzo_options = args.get('pezzo_options', None)
        native_options = {**args}
        native_options.pop('pezzo_prompt', None)
        native_options.pop('pezzo_options', None)

        managed_messages = []

        if pezzo_prompt:
            if pezzo_prompt.content.get('prompt'):
                managed_messages = [{'role': 'user', 'content': pezzo_prompt.content['prompt']}]
            if pezzo_prompt.content.get('messages'):
                managed_messages = pezzo_prompt.content['messages']

        request_body = {
            'messages': managed_messages,
            **native_options,
            **(pezzo_prompt.settings if pezzo_prompt else {}),
        }

        if pezzo_options and pezzo_options.get('variables'):
            messages = interpolate_variables_recursively(request_body['messages'], pezzo_options['variables'])
            request_body['messages'] = messages

        response, error, report_payload = None, None, None

        base_metadata = {
            'environment': self.pezzo.options.environment,
            'provider': 'OpenAI',
            'type': 'ChatCompletion',
            'client': 'pezzo-python',
            'clientVersion': get_client_version(),
        }

        request_timestamp = get_timestamp()
        merged_metadata = merge(base_metadata, pezzo_prompt.metadata if pezzo_prompt and pezzo_prompt.metadata else merge(base_metadata))

        base_report = {
            'metadata': merged_metadata,
            'cacheEnabled': False,
            'cacheHit': None,
            'request': {
                'timestamp': request_timestamp,
                'body': request_body
            }
        }

        if (pezzo_options and pezzo_options.get('properties')):
            base_report['properties'] = pezzo_options['properties']

        if pezzo_options and pezzo_options.get('cache'):
            base_report['cacheEnabled'] = True
            cached_request = self.pezzo.fetch_cached_request(request_body)

            base_report['cacheHit'] = cached_request['hit'];

            if cached_request['hit']:
                base_report['cacheHit'] = True
                response = {
                    **cached_request['data'],
                    'usage': {
                        'prompt_tokens': 0,
                        'completion_tokens': 0,
                        'total_tokens': 0,
                    }
                }
                report_payload = {
                    **base_report,
                    'response': {
                        'timestamp': request_timestamp,
                        'body': response,
                        'status': 200,
                    }
                }

        if not (pezzo_options and pezzo_options.get('cache') and base_report['cacheHit']):
            try:
                response = self.openai.ChatCompletion.create(**request_body)
                report_payload = {
                    **base_report,
                    'response': {
                        'timestamp': get_timestamp(),
                        'body': response,
                        'status': 200,
                    }
                }
            except Exception as err:
                error = err
                report_payload = {
                    **base_report,
                    'response': {
                        'timestamp': get_timestamp(),
                        'body': err.json_body,
                        'status': err.http_status,
                    }
                }

        should_write_to_cache = pezzo_options and pezzo_options.get('cache') and report_payload and report_payload['cacheHit'] == False and report_payload['response']['status'] == 200
        
        self.pezzo.report_prompt_execution(report_payload)

        if should_write_to_cache and error == None:
            self.pezzo.cache_request(request_body, response)

        if error:
            raise error

        return response

pezzo_open_ai = PezzoOpenAI()


import toml
import os

def main():
    current_path = os.path.dirname(os.path.realpath(__file__))
    pyproject_path = os.path.join(current_path, "../../pyproject.toml")
    data = toml.load(pyproject_path)
    version = data["tool"]["poetry"]["version"]
    
    # save version to "version.txt file"
    version_path = os.path.join(current_path, "../version.txt")
    with open(version_path, "w") as f:
        f.write(version)